@@ -32,6 +32,209 @@
 
 #define INT_MAX_DIGITS 8
 
+#define USEDIGIT guess = guess * 10 + (c - '0')
+
+static INLINE SVPTR
+PREFIX(number2) (parser_t * parser)
+{
+    /* End marker for strtod/strtol. */
+
+    char * end;
+
+    /* Start marker for strtod/strtol. */
+
+    char * start;
+
+    /* A guess for integer numbers. */
+
+    int guess;
+
+    /* The parsed character itself, the cause of our motion. */
+
+    char c;
+
+    /* If it has exp or dot in it. */
+
+    double d;
+
+    /* Negative number. */
+
+    int minus;
+
+    parser->end--;
+    start = parser->end;
+
+#define FAILNUMBER(err)				\
+    parser->bad_byte = parser->end - 1;		\
+    parser->error = json_error_ ## err;		\
+    parser->bad_type = json_number;		\
+    parser->bad_beginning = start;		\
+    failbadinput (parser)
+
+#define NUMBEREND				\
+         WHITESPACE:				\
+    case ']':					\
+    case '}':					\
+    case ','
+
+#define XNUMBEREND (XCOMMA|XWHITESPACE|parser->end_expected)
+
+ number_start:
+
+    minus = 0;
+
+    switch (NEXTBYTE) {
+    case DIGIT19:
+	guess = c - '0';
+	goto leading_digit19;
+    case '0':
+	goto leading_zero;
+    case '-':
+	minus = 1;
+	goto leading_minus;
+    default:
+	parser->expected = XDIGIT | XMINUS;
+	FAILNUMBER (unexpected_character);
+    }
+
+ leading_digit19:
+
+    switch (NEXTBYTE) {
+    case DIGIT:
+	USEDIGIT;
+	goto leading_digit19;
+    case '.':
+	goto dot;
+    case 'e':
+    case 'E':
+	goto exp;
+    case NUMBEREND:
+        goto int_number_end;
+    default:
+	parser->expected = XDIGIT | XDOT | XEXPONENTIAL | XNUMBEREND;
+	FAILNUMBER (unexpected_character);
+    }
+
+ leading_zero:
+    switch (NEXTBYTE) {
+    case '.':
+	/* "0." */
+	goto dot;
+    case 'e':
+    case 'E':
+	/* "0e" */
+	goto exp;
+    case NUMBEREND:
+	/* "0" */
+        goto int_number_end;
+    default:
+	parser->expected = XDOT | XEXPONENTIAL | XNUMBEREND;
+	FAILNUMBER (unexpected_character);
+    }
+
+ leading_minus:
+    switch (NEXTBYTE) {
+    case DIGIT19:
+	USEDIGIT;
+	goto leading_digit19;
+    case '0':
+	goto leading_zero;
+    default:
+	parser->expected = XDIGIT;
+	FAILNUMBER (unexpected_character);
+    }
+
+    /* Things like "5." are not allowed so there is no NUMBEREND
+       here. */
+
+ dot:
+    switch (NEXTBYTE) {
+    case DIGIT:
+	goto dot_digits;
+    default:
+	parser->expected = XDIGIT;
+	FAILNUMBER (unexpected_character);
+    }
+
+    /* We have as much as 5.5 so we can stop. */
+
+ dot_digits:
+    switch (NEXTBYTE) {
+    case DIGIT:
+	goto dot_digits;
+    case 'e':
+    case 'E':
+	goto exp;
+    case NUMBEREND:
+        goto exp_number_end;
+    default:
+	parser->expected = XDIGIT | XNUMBEREND | XEXPONENTIAL;
+	FAILNUMBER (unexpected_character);
+    }
+
+    /* Things like "10E" are not allowed so there is no NUMBEREND
+       here. */
+
+ exp:
+    switch (NEXTBYTE) {
+    case '-':
+    case '+':
+    case DIGIT:
+	goto exp_digits;
+    default:
+	parser->expected = XDIGIT | XMINUS | XPLUS;
+	FAILNUMBER (unexpected_character);
+    }
+
+    /* We have as much as "3.0e1" or similar. */
+
+ exp_digits:
+    switch (NEXTBYTE) {
+    case DIGIT:
+	goto exp_digits;
+    case NUMBEREND:
+        goto exp_number_end;
+    default:
+	parser->expected = XDIGIT | XNUMBEREND;
+	FAILNUMBER (unexpected_character);
+    }
+
+ exp_number_end:
+    parser->end--;
+    d = strtod (start, & end);
+    if (end == parser->end) {
+	RETURNAGAIN (newSVnv (d));
+    }
+    else {
+	goto string_number_end;
+    }
+
+ int_number_end:
+
+    parser->end--;
+    if (parser->end - start < INT_MAX_DIGITS + minus) {
+	if (minus) {
+	    guess = -guess;
+	}
+	RETURNAGAIN (newSViv (guess));
+    }
+    else {
+	goto string_number_end;
+    }
+
+string_number_end:
+
+    /* We could not convert this number using a number conversion
+       routine, so we are going to convert it to a string.  This might
+       happen with ridiculously long numbers or something. The JSON
+       standard doesn't explicitly disallow integers with a million
+       digits. */
+
+    RETURNAGAIN (newSVpv (start, parser->end - start));
+}
+
+#if 0
+
 /* Turn a number into an SV. */
 
 static INLINE SVPTR
@@ -43,6 +246,10 @@ PREFIX(number) (parser_t * parser)
 
     int dot;
 
+    /* Set to 1 if we saw digits after ".". */
+
+    int dotdigits;
+
     /* Set to 1 if we saw "e" or "E". */
 
     int exp;
@@ -65,6 +272,8 @@ PREFIX(number) (parser_t * parser)
 
     int newexp;
 
+    int expdigits;
+
     /* End marker for strtod/strtol. */
 
     char * end;
@@ -84,6 +293,7 @@ PREFIX(number) (parser_t * parser)
     /* Set all the flags to zero. */
 
     dot = 0;
+    dotdigits = 0;
     exp = 0;
     plus = 0;
     zero = 0;
@@ -92,6 +302,7 @@ PREFIX(number) (parser_t * parser)
     guess = 0;
     zero = 0;
     newexp = 0;
+    expdigits = 0;
 
     parser->end--;
     start = parser->end;
@@ -141,7 +352,7 @@ PREFIX(number) (parser_t * parser)
     case '-':
 
 	if (exp) {
-	    if (expminus) {
+	    if (expminus || ! newexp) {
 		parser->expected = XDIGIT;
 		FAILNUMBER (unexpected_character);
 	    }
@@ -150,7 +361,8 @@ PREFIX(number) (parser_t * parser)
 	    goto number_start;
 	}
 	else {
-	    if (minus) {
+	    /* Use "guess" as a flag that we have seen a digit. */
+	    if (minus || guess || zero) {
 		parser->expected = XDIGIT;
 		FAILNUMBER (unexpected_character);
 	    }
@@ -161,6 +373,14 @@ PREFIX(number) (parser_t * parser)
     case 'e':
     case 'E':
 
+	if (! zero && ! guess && ! dot) {
+	    parser->expected = XDIGIT;
+	    FAILNUMBER (unexpected_character);
+	}
+	if (dot && ! dotdigits) {
+	    parser->expected = XDIGIT;
+	    FAILNUMBER (unexpected_character);
+	}
 	if (exp) {
 	    parser->expected = XDIGIT;
 	    if (newexp) {
@@ -174,11 +394,19 @@ PREFIX(number) (parser_t * parser)
 
     case '0':
 
+	if (zero && ! dot && ! exp) {
+	    parser->expected = XDOT | XCOMMA | XWHITESPACE | XEXPONENTIAL | XOBJECT_END | XARRAY_END;
+	    FAILNUMBER (unexpected_character);
+	}
 	if (exp) {
 	    newexp = 0;
+	    expdigits = 1;
 	}
 	else {
-	    if (! dot) {
+	    if (dot) {
+		dotdigits = 1;
+	    }
+	    else {
 		if (guess) {
 		    guess = 10 * guess;
 		}
@@ -191,37 +419,110 @@ PREFIX(number) (parser_t * parser)
 
     case DIGIT19:
 
-	if (! dot && ! exp) {
-	    if (zero) {
-		/* "Leading zeros are not allowed." (from section 2.4
-		   of JSON standard.) */
-		FAILNUMBER (leading_zero);
+	if (zero && ! dot && ! exp) {
+	    parser->expected = XDOT | XCOMMA | XWHITESPACE | XOBJECT_END | XARRAY_END;
+	    FAILNUMBER (unexpected_character);
+	}
+	if (exp) {
+	    newexp = 0;
+	    expdigits = 1;
+	}
+	else {
+	    if (dot) {
+		dotdigits = 1;
 	    }
-	    guess = 10 * guess + (c - '0');
+	    else {
+		guess = 10 * guess + (c - '0');
+	    }
 	}
-	newexp = 0;
 	goto number_start;
 
-    case '\0':
-	if (STRINGEND) {
-	    FAILNUMBER (unexpected_end_of_input);
+    case WHITESPACE:
+    case ',':
+    case ']':
+    case '}':
+
+	if (minus && ! guess) {
+	    parser->expected = XDIGIT;
+	    FAILNUMBER (unexpected_character);
 	}
 
-	/* Fallthrough. */
+	/* JSON RFC: "The E and optional sign are followed by one or
+	   more digits." So if we have an E and no digits or anything
+	   else, barf. */
 
-    default:
+	if ((exp && ! expdigits) ||
+	    (dot && ! dotdigits)) {
+	    parser->expected = XDIGIT;
+	    if (newexp) {
+		parser->expected |= XPLUS | XMINUS;
+	    }
+	    FAILNUMBER (unexpected_character);
+	}
 
-	/* We read a byte which wasn't part of a number, so push it
-	   back on the byte stack for the next thing to work out what
-	   to do with it. */
+	/* We read a byte which wasn't part of a number, so reject it
+	   and let the next thing to work out what to do with it. */
 
 	parser->end--;
 	break;
+
+    default:
+
+	if (minus && ! guess) {
+	    parser->expected = XDIGIT;
+	    FAILNUMBER (unexpected_character);
+	}
+
+	/* JSON RFC: "The E and optional sign are followed by one or
+	   more digits." So if we have an E and no digits or anything
+	   else, barf. */
+
+	if ((exp && ! expdigits) || (dot && ! dotdigits)) {
+	    parser->expected = XDIGIT;
+	    if (newexp) {
+		parser->expected |= XPLUS | XMINUS;
+	    }
+	    FAILNUMBER (unexpected_character);
+	}
+	else {
+	    /* This looks a bit strange but we can accept a comma or
+	       whitespace or the end of an object or array as the end of a
+	       number. */
+	    parser->expected = XCOMMA | XWHITESPACE;
+	    if (zero) {
+		if (dot) {
+		    parser->expected |= XDIGIT;
+		}
+		else {
+		    if (exp) {
+			/* 0e0 is legal */
+			parser->expected |= XDIGIT;
+		    }
+		    else {
+			/* 0e is legal */
+			parser->expected |= XEXPONENTIAL;
+		    }		    
+		}
+	    }
+	    else {
+		parser->expected |= XDIGIT;
+	    }
+	    if ((guess || dotdigits) && ! exp) {
+		parser->expected |= XEXPONENTIAL;
+	    }
+	    if (! exp && ! dot && (zero || guess)) {
+		parser->expected |= XDOT;
+	    }
+	    if (newexp) {
+		parser->expected |= XPLUS | XMINUS;
+	    }
+	    FAILNUMBER (unexpected_character);
+	}
     }
 
     if (dot || exp) {
 
-	/* This is a floating point number, so we sent it to
+	/* This is a floating point number, so we send it to
 	   "strtod". */
 
 	double d;
@@ -229,6 +530,12 @@ PREFIX(number) (parser_t * parser)
 	if (end == parser->end) {
 	    RETURNAGAIN (newSVnv (d));
 	}
+	else {
+	    /* Oops, Marc Lehmann was right all along. */
+	    failbug (__FILE__, __LINE__, parser,
+		     "unable to parse '%.*s' with strtod",
+		     parser->end - start, start);
+	}
     }
     else {
 	/* If the number has less than INT_MAX_DIGITS, we guess that
@@ -253,6 +560,8 @@ PREFIX(number) (parser_t * parser)
     RETURNAGAIN (newSVpv (start, parser->end - start));
 }
 
+#endif /* 0 */
+
 static SVPTR
 PREFIX(string) (parser_t * parser)
 {
@@ -267,7 +576,7 @@ PREFIX(string) (parser_t * parser)
     len = 0;
 
     /* First of all, we examine the string to work out how long it is
-       and to look for escapes. If we find them, we go to "bad_boys"
+       and to look for escapes. If we find them, we go to "contains_escapes"
        and go back and do all the hard work of converting the escapes
        into the right things. If we don't find any escapes, we just
        use "start" and "len" and copy the string from inside
@@ -279,7 +588,7 @@ PREFIX(string) (parser_t * parser)
 	case '"':
 	    goto string_end;
 	case '\\':
-	    goto bad_boys;
+	    goto contains_escapes;
 	case BADBYTES:
 	    ILLEGALBYTE;
 	default:
@@ -298,7 +607,7 @@ PREFIX(string) (parser_t * parser)
 #endif
     goto string_done;
 
- bad_boys:
+ contains_escapes:
 
     parser->end = start;
 
@@ -319,74 +628,84 @@ PREFIX(string) (parser_t * parser)
     RETURNAGAIN (string);
 }
 
-/* JSON literals. */
+#define FAILLITERAL(c)					\
+    parser->expected = XIN_LITERAL;			\
+    parser->literal_char = c;				\
+    parser->bad_beginning = start;			\
+    parser->error = json_error_unexpected_character;	\
+    parser->bad_type = json_literal;			\
+    parser->bad_byte = parser->end - 1;			\
+    failbadinput (parser)
 
 static SVPTR
-PREFIX(literal) (parser_t * parser, char c)
+PREFIX(literal_true) (parser_t * parser)
 {
     char * start;
     start = parser->end - 1;
-    switch (c) {
-    case 't':
-	if (* parser->end++ == 'r'
-	    &&
-	    * parser->end++ == 'u'
-	    &&
-	    * parser->end++ == 'e') {
+    if (* parser->end++ == 'r') {
+	if (* parser->end++ == 'u') {
+	    if (* parser->end++ == 'e') {
 #ifdef PERLING
-	    SvREFCNT_inc (json_true);
+		SvREFCNT_inc (json_true);
 #endif
-	    RETURNAGAIN (json_true);
+		RETURNAGAIN (json_true);
+	    }
+	    FAILLITERAL('e');
 	}
-	break;
-
-    case 'n':
-	if (* parser->end++ == 'u'
-	    &&
-	    * parser->end++ == 'l'
-	    &&
-	    * parser->end++ == 'l') {
-#ifdef PERLING
-	    SvREFCNT_inc (json_null);
-#endif
-	    RETURNAGAIN (json_null);
-	}
-	break;
-
-    case 'f':
-	if (* parser->end++ == 'a'
-	    &&
-	    * parser->end++ == 'l'
-	    &&
-	    * parser->end++ == 's'
-	    &&
-	    * parser->end++ == 'e') {
-#ifdef PERLING
-	    SvREFCNT_inc (json_false);
-#endif
-	    RETURNAGAIN (json_false);
-	}
-	break;
-
-    default:
-
-	/* Because we have already checked that the input starts with
-	   either "t", "n" or "f", arriving here indicates a code
-	   failure rather than the input being wrong. */
-
-	failbug (__FILE__, __LINE__, parser,
-		 "Attempt to make a literal starting with '%02X'", c); 
+	FAILLITERAL('u');
     }
+    FAILLITERAL('r');
 
-    /* The bad character causing the failure is at "parser->end - 1"
-       because we didn't update "c" in the above switches. */
+    /* Unreached, shut up compiler warnings. */
 
-    parser->bad_byte = parser->end - 1;
-    parser->bad_type = json_literal;
-    parser->error = json_error_unexpected_character;
-    parser->bad_beginning = start;
-    parser->expected = XLITERAL;
-    failbadinput (parser); 
+    RETURNAGAIN (& PL_sv_undef);
+}
+
+static SVPTR
+PREFIX(literal_false) (parser_t * parser)
+{
+    char * start;
+    start = parser->end - 1;
+    if (* parser->end++ == 'a') {
+	if (* parser->end++ == 'l') {
+	    if (* parser->end++ == 's') {
+		if (* parser->end++ == 'e') {
+#ifdef PERLING
+		    SvREFCNT_inc (json_false);
+#endif
+		    RETURNAGAIN (json_false);
+		}
+		FAILLITERAL('e');
+	    }
+	    FAILLITERAL('s');
+	}
+	FAILLITERAL('l');
+    }
+    FAILLITERAL('a');
+
+    /* Unreached, shut up compiler warnings. */
+
+    RETURNAGAIN (& PL_sv_undef);
+}
+
+static SVPTR
+PREFIX(literal_null) (parser_t * parser)
+{
+    char * start;
+    start = parser->end - 1;
+    if (* parser->end++ == 'u') {
+	if (* parser->end++ == 'l') {
+	    if (* parser->end++ == 'l') {
+#ifdef PERLING
+		SvREFCNT_inc (json_null);
+#endif
+		RETURNAGAIN (json_null);
+	    }
+	    FAILLITERAL('l');
+	}
+	FAILLITERAL('l');
+    }
+    FAILLITERAL('u');
 
     /* Unreached, shut up compiler warnings. */
 
@@ -398,7 +717,7 @@ static SVPTR PREFIX(object) (parser_t * parser);
 /* Given one character, decide what to do next. This goes in the
    switch statement in both "object ()" and "array ()". */
 
-#define PARSE(start)				\
+#define PARSE(start,expected)			\
 						\
  case WHITESPACE:				\
  goto start;					\
@@ -409,7 +728,8 @@ static SVPTR PREFIX(object) (parser_t * parser);
 						\
  case '-':					\
  case DIGIT:					\
- SETVALUE PREFIX(number) (parser);		\
+ parser->end_expected = expected;	        \
+ SETVALUE PREFIX(number2) (parser);		\
  break;						\
 						\
  case '{':					\
@@ -421,9 +741,15 @@ static SVPTR PREFIX(object) (parser_t * parser);
  break;						\
 						\
  case 'f':					\
+ SETVALUE PREFIX(literal_false) (parser);	\
+ break;			                        \
+						\
  case 'n':					\
+ SETVALUE PREFIX(literal_null) (parser);	\
+ break;			                        \
+						\
  case 't':					\
- SETVALUE PREFIX(literal) (parser, c);	        \
+ SETVALUE PREFIX(literal_true) (parser);	\
  break
 
 /* Check for illegal comma at the end of a hash/array. */
@@ -454,46 +780,32 @@ static SVPTR
 PREFIX(array) (parser_t * parser)
 {
     char c;
+    char * start;
 #ifdef PERLING
     AV * av;
     SV * value;
 #endif
-    /* Have we seen at least one value in the array, so that commas
-       are legal? */
-    int comma;
-    char * start;
 
-    comma = 0;
 #ifdef PERLING
     av = newAV ();
 #endif
     start = parser->end - 1;
 
-    /* We are either at the start of the array, just after "[", or we
-       have seen at least one value, so just after ",". */
-
  array_start:
 
     switch (NEXTBYTE) {
 
-	PARSE(array_start);
+	BPUB (XARRAY_END);
+	PARSE (array_start,XARRAY_END);
 
     case ']':
-	CHECKCOMMA(json_array);
-	/* In legal JSON, this should only be reached for an empty
-	   array. */
 	goto array_end;
 
-    case ',':
-	parser->expected = VALUE_START | XWHITESPACE | XARRAY_END;
-	FAILARRAY(unexpected_character);
-
     default:
-	parser->expected = XARRAY_END | VALUE_START;
+	parser->expected = VALUE_START | XWHITESPACE | XARRAY_END;
 	FAILARRAY(unexpected_character);
     }
 
-    comma = 1;
 #ifdef PERLING
     av_push (av, value);
 #endif
@@ -508,7 +820,7 @@ PREFIX(array) (parser_t * parser)
 	goto array_middle;
 
     case ',':
-	goto array_start;
+	goto array_next;
 
     case ']':
 	/* Array with at least one element. */
@@ -520,6 +832,23 @@ PREFIX(array) (parser_t * parser)
 	FAILARRAY(unexpected_character);
     }
 
+ array_next:
+
+    switch (NEXTBYTE) {
+
+	PARSE(array_next, XARRAY_END);
+
+    default:
+	parser->expected = VALUE_START | XWHITESPACE;
+	FAILARRAY(unexpected_character);
+    }
+
+#ifdef PERLING
+    av_push (av, value);
+#endif
+
+    goto array_middle;
+
  array_end:
 
     RETURNAGAIN (newRV_noinc ((SV *) av));
@@ -545,10 +874,6 @@ PREFIX(object) (parser_t * parser)
     SV * value;
 #endif
     string_t key;
-    /* "middle" is true if we have seen ":", until the next comma. */
-    int middle;
-    /* "comma" is true if we have seen ",", until the next key. */
-    int comma;
     /* This is set to -1 if we want a Unicode key. See "perldoc
        perlapi" under "hv_store". */
     int uniflag;
@@ -565,11 +890,6 @@ PREFIX(object) (parser_t * parser)
 	/* Keys are not unicode. */
 	uniflag = 1;
     }
-    
-    /* We have not seen a ":" yet. */
-    middle = 0;
-    /* We have not seen a "," yet. */
-    comma = 0;
 
 #ifdef PERLING
     hv = newHV ();
@@ -578,62 +898,52 @@ PREFIX(object) (parser_t * parser)
  hash_start:
 
     switch (NEXTBYTE) {
-
     case WHITESPACE:
 	goto hash_start;
-
-	/* Unreachable */
-
     case '}':
-	CHECKCOMMA(json_object);
 	goto hash_end;
-
-	/* Unreachable */
-
     case '"':
-	if (middle) {
-	    FAILOBJECT(unexpected_character);
-	}
-	else {
-	    comma = 0;
-	    get_key_string (parser, & key);
-	    goto hash_next;
-	}
+	get_key_string (parser, & key);
+	goto hash_next;
+    default:
+	parser->expected = XWHITESPACE | XSTRING_START | XOBJECT_END;
+	FAILOBJECT(unexpected_character);
+    }
 
-	/* Unreachable */
+ hash_middle:
 
+    switch (NEXTBYTE) {
+    case WHITESPACE:
+	goto hash_middle;
+    case '}':
+	goto hash_end;
     case ',':
-	if (middle) {
-	    middle = 0;
-	    comma = 1;
-	    goto hash_start;
-	}
-	else {
-	    parser->expected = XWHITESPACE | XSTRING_START;
-	    FAILOBJECT(unexpected_character);
-	}
+	goto hash_key;
+    default:
+	parser->expected = XWHITESPACE | XCOMMA | XOBJECT_END;
+	FAILOBJECT(unexpected_character);
+    }
 
-	/* Unreachable */
+ hash_key:
 
+    switch (NEXTBYTE) {
+    case WHITESPACE:
+	goto hash_key;
+    case '"':
+	get_key_string (parser, & key);
+	goto hash_next;
     default:
 	parser->expected = XWHITESPACE | XSTRING_START;
-	if (middle) {
-	    parser->expected |= XCOMMA;
-	}
 	FAILOBJECT(unexpected_character);
     }
 
  hash_next:
 
     switch (NEXTBYTE) {
-
     case WHITESPACE:
 	goto hash_next;
-
     case ':':
-	middle = 1;
 	goto hash_value;
-
     default:
 	parser->expected = XWHITESPACE | XVALUE_SEPARATOR;
 	FAILOBJECT(unexpected_character);
@@ -644,15 +954,13 @@ PREFIX(object) (parser_t * parser)
  hash_value:
 
     switch (NEXTBYTE) {
-
-	PARSE(hash_value);
-
+	PARSE(hash_value, XOBJECT_END);
     default:
 	parser->expected = XWHITESPACE | VALUE_START;
 	FAILOBJECT(unexpected_character);
     }
 
-    if (key.bad_boys) {
+    if (key.contains_escapes) {
 	int klen;
 	klen = resolve_string (parser, & key);
 #ifdef PERLING
@@ -664,8 +972,7 @@ PREFIX(object) (parser_t * parser)
 	(void) hv_store (hv, key.start, key.length * uniflag, value, 0);
 #endif
     }
-
-    goto hash_start;
+    goto hash_middle;
 
  hash_end:
 
